{
--------------------------------------------------------------------------------
This unit contains a class named TCheckListBoxEx. In addition to the standard
TCheckListBox class provided by the VCL it adds a CheckBoxCount property, which
makes it possible to add mulitple checkboxes for an item.
It's not based on TCheckListBox itself, but is a descendant of TCustomListBox.
This is because some very important members are declared private in
TCheckListBox, which means they cannot be overriden.
--------------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance with
the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

Alternatively, the contents of this file may be used under the terms of the
GNU General Public License Version 2 or later (the "GPL"), in which case
the provisions of the GPL are applicable instead of those above.
If you wish to allow use of your version of this file only under the terms
of the GPL and not to allow others to use your version of this file
under the MPL, indicate your decision by deleting the provisions above and
replace them with the notice and other provisions required by the GPL.
If you do not delete the provisions above, a recipient may use your version
of this file under either the MPL or the GPL.
--------------------------------------------------------------------------------
This unit is based on the unit "CheckLst" contributed with the VCL by Borland.
Portions created by Borland are Copyright by Borland. All Rights Reserved.

Contributor:  Michael Elsdrfer <michael@elsdoerfer.net>

You may retrieve the latest version of this file at http://www.elsdoerfer.net
--------------------------------------------------------------------------------
}

unit ExtCheckListBox;

interface

uses Windows, Messages, Classes, Graphics, StdCtrls, Controls, SysUtils;

type
  TOnClickCheckEvent = procedure(Sender: TObject; ItemIndex, CheckIndex: Integer;
    NewState: TCheckBoxState) of object;

  TCheckListBoxEx = class(TCustomListBox)
  private
    FCheckBoxCount: Integer;
    FAllowGrayed: Boolean;
    FFlat: Boolean;
    FHeaderColor: TColor;
    FHeaderBackgroundColor: TColor;
    FOnClickCheck: TOnClickCheckEvent;
    procedure SetCheckBoxCount(const Value: Integer);
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMDestroy(var Msg : TWMDestroy);message WM_DESTROY;
    function GetChecked(ItemIndex, CheckIndex: Integer): Boolean;
    procedure SetChecked(ItemIndex, CheckIndex: Integer;
      const Value: Boolean);
    function GetItemEnabled(Index: Integer): Boolean;
    function GetState(ItemIndex, CheckIndex: Integer): TCheckBoxState;
    procedure SetItemEnabled(Index: Integer; const Value: Boolean);
    procedure SetState(ItemIndex, CheckIndex: Integer;
      const Value: TCheckBoxState);
    function GetHeader(Index: Integer): Boolean;
    procedure SetHeader(Index: Integer; const Value: Boolean);
    procedure SetFlat(const Value: Boolean);
    procedure SetHeaderBackgroundColor(const Value: TColor);
    procedure SetHeaderColor(const Value: TColor);
  protected
    procedure DrawCheck(R: TRect; AState: TCheckBoxState; AEnabled: Boolean);
    procedure DrawItem(Index: Integer; Rect: TRect;
      State: TOwnerDrawState); override;
    procedure InvalidateCheck(ItemIndex: Integer; CheckIndex: Integer = -1);
    function InternalGetItemData(Index: Integer): Longint; override;
    procedure InternalSetItemData(Index: Integer; AData: Longint); override;
    procedure SetItemData(Index: Integer; AData: LongInt); override;
    function GetItemData(Index: Integer): LongInt; override;
    function CreateWrapper(Index: Integer): TObject;
    function ExtractWrapper(Index: Integer): TObject;
    function GetWrapper(Index: Integer): TObject;
    function HaveWrapper(Index: Integer): Boolean;
    procedure ToggleClickCheck(ItemIndex, CheckIndex: Integer);
    procedure KeyPress(var Key: Char); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure ResetContent; override;
    procedure DeleteString(Index: Integer); override;
    procedure ClickCheck(ItemIndex, CheckIndex: Integer; NewState: TCheckBoxState); dynamic;
    procedure CreateParams(var Params: TCreateParams); override;
    function GetCheckWidth: Integer;
    procedure ResetItemHeight;
    procedure CreateWnd; override;
    function WidthOfCheckBoxes: Integer;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property Checked[ItemIndex, CheckIndex: Integer]: Boolean read GetChecked write SetChecked;
    property ItemEnabled[Index: Integer]: Boolean read GetItemEnabled write SetItemEnabled;
    property State[ItemIndex, CheckIndex: Integer]: TCheckBoxState read GetState write SetState;
    property Header[Index: Integer]: Boolean read GetHeader write SetHeader;
  published
    property OnClickCheck: TOnClickCheckEvent read FOnClickCheck write FOnClickCheck;
    property Align;
    property AllowGrayed: Boolean read FAllowGrayed write FAllowGrayed default False;
    property Anchors;
    property AutoComplete;
    property BevelEdges;
    property BevelInner;
    property BevelOuter;
    property BevelKind;
    property BevelWidth;
    property BiDiMode;
    property BorderStyle;
    property CheckBoxCount: Integer read FCheckBoxCount write SetCheckBoxCount;    
    property Color;
    property Columns;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Flat: Boolean read FFlat write SetFlat default True;
    property Font;
    property HeaderColor: TColor read FHeaderColor write SetHeaderColor default clInfoText;
    property HeaderBackgroundColor: TColor read FHeaderBackgroundColor write SetHeaderBackgroundColor default clInfoBk;
    property ImeMode;
    property ImeName;
    property IntegralHeight;
    property ItemHeight;
    property Items;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property Style;
    property TabOrder;
    property TabStop;
    property TabWidth;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnData;
    property OnDataFind;
    property OnDataObject;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;  

implementation

uses Consts, RTLConsts, Themes;

var
  FCheckWidth, FCheckHeight: Integer;

procedure GetCheckSize;
begin
  with TBitmap.Create do
    try
      Handle := LoadBitmap(0, PChar(OBM_CHECKBOXES));
      FCheckWidth := Width div 4;
      FCheckHeight := Height div 3;
    finally
      Free;
    end;
end;

procedure Register;
Begin
  RegisterComponents('elsdoerfer.net', [TCheckListBoxEx]);
end;

type
  TCheckBoxExItemState = array of TCheckBoxState;

  TCheckListBoxExDataWrapper = class
  private
    FData: LongInt;
    FStates: TCheckBoxExItemState;
    FDisabled: Boolean;
    FHeader: Boolean;
    function GetChecked(CheckIndex: Integer): Boolean;
    procedure SetChecked(CheckIndex: Integer; const Value: Boolean);
    procedure SetStates(Index: Integer; const Value: TCheckBoxState);
    function GetStates(Index: Integer): TCheckBoxState;
  protected
    // States for each CheckBox of an item are stored in an array - use
    // these interal methods to read and write the states, though they dynamically
    // enlarge the array as needed. As a result, we have not to manage the sizes
    // of the state arrays of each item when TCheckListBoxEx.CheckBoxCount is changed.
    procedure InternalSetState(Index: Integer; Value: TCheckBoxState);
    function InternalGetState(Index: Integer): TCheckBoxState;
  public
    constructor Create;
    destructor Destroy; override;
        
    class function GetDefaultState: TCheckBoxState;
    property Checked[CheckIndex: Integer]: Boolean read GetChecked write SetChecked;
    property States[Index: Integer]: TCheckBoxState read GetStates write SetStates;
    property Disabled: Boolean read FDisabled write FDisabled;
    property Header: Boolean read FHeader write FHeader;
  end;  

{ TCheckListBoxEx }

procedure TCheckListBoxEx.ClickCheck(ItemIndex, CheckIndex: Integer;
  NewState: TCheckBoxState);
begin
  if Assigned(FOnClickCheck) then FOnClickCheck(Self, ItemIndex, CheckIndex, NewState);
end;

procedure TCheckListBoxEx.CMFontChanged(var Message: TMessage);
begin
  inherited;
  ResetItemHeight;
end;

procedure TCheckListBoxEx.CNDrawItem(var Message: TWMDrawItem);
begin
  if Items.Count = 0 then exit;
  with Message.DrawItemStruct^ do
    if not Header[itemID] then
      if not UseRightToLeftAlignment then
        rcItem.Left := rcItem.Left + WidthOfCheckBoxes
      else rcItem.Right := rcItem.Right - WidthOfCheckBoxes;
  inherited;
end;

constructor TCheckListBoxEx.Create(AOwner: TComponent);
begin
  inherited;
  FCheckBoxCount := 1;
  Flat := True;
  FHeaderColor := clInfoText;
  FHeaderBackgroundColor := clInfoBk;
end;

procedure TCheckListBoxEx.CreateParams(var Params: TCreateParams);
begin
  inherited;
  with Params do
    if Style and (LBS_OWNERDRAWFIXED or LBS_OWNERDRAWVARIABLE) = 0 then
      Style := Style or LBS_OWNERDRAWFIXED;
end;

procedure TCheckListBoxEx.CreateWnd;
begin
  inherited;
  ResetItemHeight;
end;

function TCheckListBoxEx.CreateWrapper(Index: Integer): TObject;
begin
  Result := TCheckListBoxExDataWrapper.Create;
  inherited SetItemData(Index, LongInt(Result));
end;

procedure TCheckListBoxEx.DeleteString(Index: Integer);
begin
  if HaveWrapper(Index) then
    GetWrapper(Index).Free;
  inherited;
end;

destructor TCheckListBoxEx.Destroy;
begin
  inherited;
end;

procedure TCheckListBoxEx.DrawCheck(R: TRect; AState: TCheckBoxState;
  AEnabled: Boolean);
var DrawState: Integer;
    DrawRect: TRect;
    OldBrushColor: TColor;
    OldBrushStyle: TBrushStyle;
    OldPenColor: TColor;
    Rgn, SaveRgn: HRgn;
    ElementDetails: TThemedElementDetails;
begin
  SaveRgn := 0;
  DrawRect.Left := R.Left + (R.Right - R.Left - FCheckWidth) div 2;
  DrawRect.Top := R.Top + (R.Bottom - R.Top - FCheckHeight) div 2;
  DrawRect.Right := DrawRect.Left + FCheckWidth;
  DrawRect.Bottom := DrawRect.Top + FCheckHeight;
  with Canvas do
  begin
    if Flat then
    begin
      { Remember current clipping region }
      SaveRgn := CreateRectRgn(0,0,0,0);
      GetClipRgn(Handle, SaveRgn);
      { Clip 3d-style checkbox to prevent flicker }
      with DrawRect do
        Rgn := CreateRectRgn(Left + 2, Top + 2, Right - 2, Bottom - 2);
      SelectClipRgn(Handle, Rgn);
      DeleteObject(Rgn);
    end;

   if ThemeServices.ThemesEnabled then
   begin
      case AState of
        cbChecked:
          if AEnabled then
            ElementDetails := ThemeServices.GetElementDetails(tbCheckBoxCheckedNormal)
          else
            ElementDetails := ThemeServices.GetElementDetails(tbCheckBoxCheckedDisabled);
        cbUnchecked:
          if AEnabled then
            ElementDetails := ThemeServices.GetElementDetails(tbCheckBoxUncheckedNormal)
          else
            ElementDetails := ThemeServices.GetElementDetails(tbCheckBoxUncheckedDisabled)
        else // cbGrayed
          if AEnabled then
            ElementDetails := ThemeServices.GetElementDetails(tbCheckBoxMixedNormal)
          else
            ElementDetails := ThemeServices.GetElementDetails(tbCheckBoxMixedDisabled);
      end;
      ThemeServices.DrawElement(Handle, ElementDetails, R);
    end
    else
    begin
      case AState of
        cbChecked:
          DrawState := DFCS_BUTTONCHECK or DFCS_CHECKED;
        cbUnchecked:
          DrawState := DFCS_BUTTONCHECK;
        else // cbGrayed
          DrawState := DFCS_BUTTON3STATE or DFCS_CHECKED;
      end;
      if not AEnabled then
        DrawState := DrawState or DFCS_INACTIVE;
      DrawFrameControl(Handle, DrawRect, DFC_BUTTON, DrawState);
    end;

    if Flat then
    begin
      SelectClipRgn(Handle, SaveRgn);
      DeleteObject(SaveRgn);
      { Draw flat rectangle in-place of clipped 3d checkbox above }
      OldBrushStyle := Brush.Style;
      OldBrushColor := Brush.Color;
      OldPenColor := Pen.Color;
      Brush.Style := bsClear;
      Pen.Color := clBtnShadow;
      with DrawRect do
        Rectangle(Left + 1, Top + 1, Right - 1, Bottom - 1);
      Brush.Style := OldBrushStyle;
      Brush.Color := OldBrushColor;
      Pen.Color := OldPenColor;
    end;
  end;
end;

procedure TCheckListBoxEx.DrawItem(Index: Integer; Rect: TRect;
  State: TOwnerDrawState);
var
  R: TRect;
  SaveEvent: TDrawItemEvent;
  ACheckWidth: Integer;
  Enable: Boolean;
  I: Integer;
begin
  ACheckWidth := fCheckWidth;
  if Index < Items.Count then
  begin
    R := Rect;
    Enable := Self.Enabled and ItemEnabled[Index];
    if not Header[Index] then
    begin
      // Initalize Rect
      if not UseRightToLeftAlignment then
      begin
        R.Right := Rect.Left - 2;
        R.Left := R.Right - ACheckWidth;
      end else
      begin
        R.Left := Rect.Right + 2;
        R.Right := R.Left + ACheckWidth;
      end;

      // Draw Checks
      for I := 0 to CheckBoxCount - 1 do Begin
        // Draw
        DrawCheck(R, Self.State[Index, I], Enable);

        // Calc new Rect
        if not UseRightToLeftAlignment then
        begin
          R.Right:= R.Left;
          R.Left := R.Left - ACheckWidth;
        end else
        begin
          R.Left := R.Right;
          R.Right := R.Right + ACheckWidth;
        end;
      end;
    end
    else
    begin
      Canvas.Font.Color := HeaderColor;
      Canvas.Brush.Color := HeaderBackgroundColor;
    end;
    if not Enable then
      Canvas.Font.Color := clGrayText;
  end;

  if (Style = lbStandard) and Assigned(OnDrawItem) then
  begin
    { Force lbStandard list to ignore OnDrawItem event }
    SaveEvent := OnDrawItem;
    OnDrawItem := nil;
    try
      inherited;
    finally
      OnDrawItem := SaveEvent;
    end;
  end else inherited;
end;

function TCheckListBoxEx.ExtractWrapper(Index: Integer): TObject;
begin
  Result := TCheckListBoxExDataWrapper(inherited GetItemData(Index));
  if LB_ERR = Integer(Result) then
    raise EListError.CreateResFmt(@SListIndexError, [Index]);
  if (Result <> nil) and (not (Result is TCheckListBoxExDataWrapper)) then Result := nil;
end;

function TCheckListBoxEx.GetChecked(ItemIndex,
  CheckIndex: Integer): Boolean;
begin
  if HaveWrapper(ItemIndex) then
    Result := TCheckListBoxExDataWrapper(GetWrapper(ItemIndex)).GetChecked(CheckIndex)
  else Result := False;
end;

function TCheckListBoxEx.GetCheckWidth: Integer;
begin
  Result := FCheckWidth + 2;
end;

function TCheckListBoxEx.GetHeader(Index: Integer): Boolean;
begin
  if HaveWrapper(Index) then
    Result := TCheckListBoxExDataWrapper(GetWrapper(Index)).Header
  else Result := False;
end;

function TCheckListBoxEx.GetItemData(Index: Integer): LongInt;
begin
  Result := 0;
  if HaveWrapper(Index) then
    Result := TCheckListBoxExDataWrapper(GetWrapper(Index)).FData;
end;

function TCheckListBoxEx.GetItemEnabled(Index: Integer): Boolean;
begin
  if HaveWrapper(Index) then
    Result := not TCheckListBoxExDataWrapper(GetWrapper(Index)).Disabled
  else
    Result := True;
end;

function TCheckListBoxEx.GetState(ItemIndex,
  CheckIndex: Integer): TCheckBoxState;
begin
  if HaveWrapper(ItemIndex) then
    Result := TCheckListBoxExDataWrapper(GetWrapper(ItemIndex)).States[CheckIndex]
  else
    Result := TCheckListBoxExDataWrapper.GetDefaultState;
end;

function TCheckListBoxEx.GetWrapper(Index: Integer): TObject;
begin
  Result := ExtractWrapper(Index);
  if Result = nil then
    Result := CreateWrapper(Index);
end;

function TCheckListBoxEx.HaveWrapper(Index: Integer): Boolean;
begin
  Result := ExtractWrapper(Index) <> nil;
end;

function TCheckListBoxEx.InternalGetItemData(Index: Integer): Longint;
begin
  Result := inherited GetItemData(Index);
end;

procedure TCheckListBoxEx.InternalSetItemData(Index, AData: Integer);
begin
  inherited SetItemData(Index, AData);
end;

procedure TCheckListBoxEx.InvalidateCheck(ItemIndex: Integer; CheckIndex: Integer = -1);
var R: TRect;
begin
  if not Header[ItemIndex] then
  begin
    R := ItemRect(ItemIndex);
    if not UseRightToLeftAlignment then
      if CheckIndex = -1 then R.Right := R.Left + fCheckWidth * CheckBoxCount
      else Begin
        R.Left := (CheckBoxCount - CheckIndex - 1) * fCheckWidth + R.Left;
        R.Right := R.Left + fCheckWidth
      end
    else
      if CheckIndex = -1 then R.Left := R.Right - fCheckWidth * CheckBoxCount
      else Begin
         R.Left := R.Right - (CheckBoxCount - CheckIndex - 1) * fCheckWidth;
         R.Right := R.Left + fCheckWidth
      end;
    InvalidateRect(Handle, @R, not (csOpaque in ControlStyle));
    UpdateWindow(Handle);
  end;
end;

procedure TCheckListBoxEx.KeyPress(var Key: Char);
var I: Integer;
begin
  if (Key = ' ') then
    for I  := 0 to CheckBoxCount - 1 do ToggleClickCheck(ItemIndex, I);
  inherited KeyPress(Key);
end;

procedure TCheckListBoxEx.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var Index, CheckIndex: Integer;
begin
  inherited;
  if Button = mbLeft then
  begin
    Index := ItemAtPos(Point(X,Y),True);
    if (Index <> -1) and GetItemEnabled(Index) then
      if not UseRightToLeftAlignment then
      begin
        if X - ItemRect(Index).Left < WidthOfCheckBoxes then Begin
          CheckIndex := CheckBoxCount - ((X - 2) div FCheckWidth) - 1;
          if (CheckIndex <> -1) then ToggleClickCheck(Index, CheckIndex)
        end;
      end else begin
        { TODO : TODO: Support for ReftToLight Alignment }
        Dec(X, ItemRect(Index).Right - GetCheckWidth);
        if (X > 0) and (X < GetCheckWidth) then
          ToggleClickCheck(Index, 9)
      end;
  end;
end;

procedure TCheckListBoxEx.ResetContent;
var I: Integer;
begin
  for I := 0 to Items.Count - 1 do
    if HaveWrapper(I) then
      GetWrapper(I).Free;
  inherited;
end;

procedure TCheckListBoxEx.ResetItemHeight;
begin
  if HandleAllocated and (Style = lbStandard) then
  begin
    Canvas.Font := Font;
    Perform(LB_SETITEMHEIGHT, 0, Canvas.TextHeight('Wg'));
  end;
end;

procedure TCheckListBoxEx.SetCheckBoxCount(const Value: Integer);
begin
  if FCheckBoxCount <> Value then Begin
    FCheckBoxCount := Value;
    Invalidate;
  end;
end;

procedure TCheckListBoxEx.SetChecked(ItemIndex, CheckIndex: Integer;
  const Value: Boolean);
begin
  if Value <> GetChecked(ItemIndex, CheckIndex) then
  begin
    TCheckListBoxExDataWrapper(GetWrapper(ItemIndex)).SetChecked(CheckIndex, Value);
    InvalidateCheck(ItemIndex, CheckIndex);
  end;
end;

procedure TCheckListBoxEx.SetFlat(const Value: Boolean);
begin
  if Value <> FFlat then
  begin
    FFlat := Value;
    Invalidate;
  end;
end;

procedure TCheckListBoxEx.SetHeader(Index: Integer; const Value: Boolean);
begin
  if Value <> GetHeader(Index) then
  begin
    TCheckListBoxExDataWrapper(GetWrapper(Index)).Header := Value;
    InvalidateCheck(Index);
  end;
end;

procedure TCheckListBoxEx.SetHeaderBackgroundColor(const Value: TColor);
begin
  if Value <> HeaderBackgroundColor then
  begin
    FHeaderBackgroundColor := Value;
    Invalidate;
  end;
end;

procedure TCheckListBoxEx.SetHeaderColor(const Value: TColor);
begin
  if Value <> HeaderColor then
  begin
    FHeaderColor := Value;
    Invalidate;
  end;
end;

procedure TCheckListBoxEx.SetItemData(Index, AData: Integer);
var Wrapper: TCheckListBoxExDataWrapper;
begin
  if HaveWrapper(Index) or (AData <> 0) then
  begin
    Wrapper := TCheckListBoxExDataWrapper(GetWrapper(Index));
    Wrapper.FData := AData;
  end;
end;

procedure TCheckListBoxEx.SetItemEnabled(Index: Integer;
  const Value: Boolean);
begin
  if Value <> GetItemEnabled(Index) then
  begin
    TCheckListBoxExDataWrapper(GetWrapper(Index)).Disabled := not Value;
    InvalidateCheck(Index);
  end;
end;

procedure TCheckListBoxEx.SetState(ItemIndex, CheckIndex: Integer;
  const Value: TCheckBoxState);
begin
  if Value <> GetState(ItemIndex, CheckIndex) then
  begin
    TCheckListBoxExDataWrapper(GetWrapper(ItemIndex)).States[CheckIndex] := Value;
    InvalidateCheck(ItemIndex, CheckIndex);
  end;
end;

procedure TCheckListBoxEx.ToggleClickCheck(ItemIndex, CheckIndex: Integer);
var
  State: TCheckBoxState;
begin
  if (ItemIndex >= 0) and (ItemIndex < Items.Count) and GetItemEnabled(ItemIndex) then
  begin
    State := Self.State[ItemIndex, CheckIndex];
    case State of
      cbUnchecked:
        if AllowGrayed then State := cbGrayed else State := cbChecked;
      cbChecked: State := cbUnchecked;
      cbGrayed: State := cbChecked;
    end;
    Self.State[ItemIndex, CheckIndex] := State;
    ClickCheck(ItemIndex, CheckIndex, State);
  end;
end;

function TCheckListBoxEx.WidthOfCheckBoxes: Integer;
begin
  Result := FCheckWidth * (FCheckBoxCount - 1) + GetCheckWidth + 1;
end;

procedure TCheckListBoxEx.WMDestroy(var Msg: TWMDestroy);
var i: Integer;
begin
  for i := 0 to Items.Count -1 do
    ExtractWrapper(i).Free;
  inherited;
end;

{ TCheckListBoxExDataWrapper }

constructor TCheckListBoxExDataWrapper.Create;
begin
  SetLength(fStates,0);
end;

destructor TCheckListBoxExDataWrapper.Destroy;
begin
  SetLength(fStates,0);
  fStates := nil;
  inherited;
end;

function TCheckListBoxExDataWrapper.GetChecked(CheckIndex: Integer): Boolean;
begin
  Result := States[CheckIndex] = cbChecked;
end;

class function TCheckListBoxExDataWrapper.GetDefaultState: TCheckBoxState;
begin
  Result := cbUnchecked;
end;

function TCheckListBoxExDataWrapper.GetStates(
  Index: Integer): TCheckBoxState;
begin
  Result := InternalGetState(Index)
end;

function TCheckListBoxExDataWrapper.InternalGetState(
  Index: Integer): TCheckBoxState;
begin
  if (Index > high(fStates)) then Result := GetDefaultState
  else Result := fStates[Index];
end;

procedure TCheckListBoxExDataWrapper.InternalSetState(Index: Integer;
  Value: TCheckBoxState);
begin
  // enlare array  if needed
  if (Index > high(fStates)) then SetLength(fStates, Index + 1);
  // set value
  fStates[Index] := Value;
end;

procedure TCheckListBoxExDataWrapper.SetChecked(CheckIndex: Integer;
  const Value: Boolean);
begin
  if Value then InternalSetState(CheckIndex, cbChecked)
  else InternalSetState(CheckIndex, cbUnChecked); 
end;

procedure TCheckListBoxExDataWrapper.SetStates(Index: Integer;
  const Value: TCheckBoxState);
begin
  InternalSetState(Index, Value);
end;

initialization
  GetCheckSize;

end.
